import torch
import json
from torch import nn
from torch.nn import CrossEntropyLoss
from transformers import AutoModel, AutoConfig

class SuperHFModel(AutoModel):

    @classmethod
    def set_sample_config(cls, pretrained_model_name_or_path, **kwargs):
        is_pretrained = kwargs.pop("is_pretrained", True)
        # Create the candidate net with random initialization
        candidate_hf_config = AutoConfig.from_pretrained(pretrained_model_name_or_path, **kwargs)
        candidate_hf = cls.from_config(candidate_hf_config)

        if is_pretrained:
            # Create the super net
            super_hf = cls.from_pretrained(pretrained_model_name_or_path)
            # Load pre-trained weight from super net
            candidate_hf_state_dict = candidate_hf.state_dict()
            super_hf_state_dict = super_hf.state_dict()

            new_candidate_hf_state_dict = {}
            for k in candidate_hf_state_dict:
                super_hf_state = super_hf_state_dict[k]
                candidate_hf_state = super_hf_state
                for dim, size in enumerate(candidate_hf_state_dict[k].size()):
                    candidate_hf_state = candidate_hf_state.index_select(dim, torch.tensor(range(size)))
                new_candidate_hf_state_dict[k] = candidate_hf_state
            candidate_hf.load_state_dict(new_candidate_hf_state_dict)

        return candidate_hf

    @classmethod
    def search_space_generation(cls, pretrained_model_name_or_path, **kwargs):
        hf_config = AutoConfig.from_pretrained(pretrained_model_name_or_path)
        search_space = {}
        search_space["num_hidden_layers"] = list(range(int(hf_config.num_hidden_layers/2), int(hf_config.num_hidden_layers), 1))
        search_space["num_attention_heads"] = list(range(int(hf_config.num_attention_heads/2), int(hf_config.num_attention_heads), 1))
        search_space["hidden_size"] = list(range(int(hf_config.hidden_size/2), int(hf_config.hidden_size), 16))
        for k in kwargs:
            if 'max' not in kwargs[k]:
                raise ValueError("Please specify the up bound of {} in search space".format(k))
            search_range = {"min": int(kwargs[k]['max']/2), "max": int(kwargs[k]['max']), "step": 1}
            if 'min' in kwargs[k]:
                search_range["min"] = int(kwargs[k]["min"])
            if 'step' in kwargs[k]:
                search_range["step"] = int(kwargs[k]["step"])
            search_space[k] = list(range(search_range["min"], search_range["max"], search_range["step"]))
        return search_space



import os
import ast
import time
import random
import numpy as np
import json
import shutil
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.modules import linear
from transformers import pytorch_utils

from e2eAIOK.DeNas.thirdparty.supernet_hf import SuperHFModel

LINEAR_LAYER_STRUCTURE = {"output": linear.Linear,
                            "intermediate": linear.Linear,
                            "mlp": pytorch_utils.Conv1D
                        }

ATTN_LAYER_STRUCTURE = {"selfatt": linear.Linear,
                            "att": pytorch_utils.Conv1D
                        }

def decode_arch(arches_file):
    subhf_config = None
    with open(arches_file, 'r') as fin:
        for line in fin:
            line = line.strip()
            subhf_config = ast.literal_eval(line)
    return subhf_config

def hf_populate_random_func(search_space):
    cand_tuple = dict()
    for search_key in search_space:
        cand_tuple[search_key] = random.choice(search_space[search_key]) #nosec

    cand_tuple["hidden_size"] = int(cand_tuple["hidden_size"]/cand_tuple["num_attention_heads"]) * cand_tuple["num_attention_heads"]

    return json.dumps(cand_tuple)

def hf_is_legal(cand, vis_dict, params):
    if cand not in vis_dict:
        vis_dict[cand] = {}
    info = vis_dict[cand]
    if 'visited' in info:
        return False
    cand = json.loads(cand)
    candidate_net = SuperHFModel.set_sample_config(params.supernet, **cand)
    n_parameters = sum(param.numel() for param in candidate_net.parameters())
    info['params'] = n_parameters / 10.**6
    if info['params'] > params.max_param_limits:
        return False
    if info['params'] < params.min_param_limits:
        return False
    info['visited'] = True
    return True

def hf_mutation_random_func(m_prob, s_prob, search_space, top_candidates):
    cand = random.choice(top_candidates) #nosec
    random_s = random.random() #nosec
    result_cand = json.loads(cand)
    
    if random_s < m_prob:
        # num_heads
        result_cand['num_attention_heads'] = random.choice(search_space['num_attention_heads']) #nosec
    for search_key in search_space:
        if search_key == 'num_attention_heads':
            continue
        random_s = random.random() #nosec
        if random_s < s_prob:
            result_cand[search_key] = random.choice(search_space[search_key]) #nosec

    result_cand["hidden_size"] = int(result_cand["hidden_size"]/result_cand["num_attention_heads"]) * result_cand["num_attention_heads"]

    return json.dumps(result_cand)

def hf_crossover_random_func(top_candidates):
    p1 = json.loads(random.choice(top_candidates)) #nosec
    p2 = json.loads(random.choice(top_candidates)) #nosec
    max_iters_tmp = 50
    while len(p1) != len(p2) and max_iters_tmp > 0:
        max_iters_tmp -= 1
        p1 = json.loads(random.choice(top_candidates)) #nosec
        p2 = json.loads(random.choice(top_candidates)) #nosec
    cand = {}
    for it1, it2 in zip(p1, p2):
        select_item = random.choice([p1[it1], p2[it2]]) #nosec
        cand[it1] = select_item
    cand["hidden_size"] = int(cand["hidden_size"]/cand["num_attention_heads"]) * cand["num_attention_heads"]
    return json.dumps(cand)

def get_hf_latency(model, batch_size, max_seq_length, gpu, infer_cnt):
    if gpu is None:
        device = 'cpu'
    else:
        device = 'cuda'
    
    batch = input_construtor(batch_size, max_seq_length)

    aver_time = 0.
    model.eval()

    for i in range(int(infer_cnt)):
        start = time.time()
        with torch.no_grad():
            model(**batch)
        end = time.time()
        sep = 1000 * (end - start)
        if i == 0:
            continue
        else:
            aver_time += sep / (infer_cnt - 1)
    return aver_time

def input_construtor(batch_size, resolution, domain="NLP"):
    if domain == "NLP":
        input_ids = [[9333-id] * resolution for id in range(batch_size)]
        input_masks = resolution * [1]
        input_segments = resolution * [0]
        input_ids = torch.tensor(input_ids, dtype=torch.long)
        input_masks = torch.tensor([input_masks]*batch_size, dtype=torch.long)
        input_segments = torch.tensor([input_segments]*batch_size, dtype=torch.long)
        batch = {'input_ids': input_ids, 'token_type_ids': input_segments, 'attention_mask':input_masks}
        return batch
    else:
        raise NotImplementedError


import os
import torch
from torchvision import datasets, transforms
from e2eAIOK.common.trainer.data.data_builder_cv import DataBuilderCV
from e2eAIOK.common.utils import check_mkdir
from timm.data.constants import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD
from timm.data import create_transform

class DataBuilderImageNet(DataBuilderCV):
    def __init__(self, cfg):
        super().__init__(cfg)
    
    def prepare_dataset(self):
        """
            prepare ImageNet dataset
        """
        transform = self.build_transform(is_train = True)
        train_data_folder = check_mkdir(self.cfg.train_data_path)
        self.dataset_train = datasets.ImageFolder(train_data_folder, transform=transform)

        transform = self.build_transform(is_train = False)
        val_data_folder = check_mkdir(self.cfg.val_data_path)
        self.dataset_val = datasets.ImageFolder(val_data_folder, transform=transform)
        


    def build_transform(self, is_train):
        resize_im = "input_size" in self.cfg and self.cfg.input_size > 32
        if is_train:
            # this should always dispatch to transforms_imagenet_train
            transform = create_transform(
                input_size=self.cfg.input_size,
                is_training=True,
                color_jitter=self.cfg.color_jitter,
                auto_augment=self.cfg.aa,
                interpolation=self.cfg.train_interpolation,
                re_prob=self.cfg.reprob,
                re_mode=self.cfg.remode,
                re_count=self.cfg.recount,
            )
            if not resize_im:
                # replace RandomResizedCropAndInterpolation with
                # RandomCrop
                transform.transforms[0] = transforms.RandomCrop(
                    self.cfg.input_size, padding=4)
            return transform

        t = []
        if resize_im:
            size = int((256 / 224) * self.cfg.input_size)
            t.append(
                transforms.Resize(size, interpolation=3),  # to maintain same ratio w.r.t. 224 images
            )
            t.append(transforms.CenterCrop(self.cfg.input_size))

        t.append(transforms.ToTensor())
        t.append(transforms.Normalize(IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD))
        return transforms.Compose(t)